lkml.org 
[lkml]   [2003]   [Aug]   [9]   [last100]   RSS Feed
Views: [wrap][no wrap]   [headers]  [forward] 
 
Messages in this thread
Patch in this message
/
Date
From
Subject[PATCH] 2.6.0-test3 - Watchdog patches (w83877f_wdt.c and sbc60xxwdt.c)
Hi Linus,

please do a

bk pull http://linux-watchdog.bkbits.net/linux-2.5-watchdog

This will update the following files:

drivers/char/watchdog/sbc60xxwdt.c | 4
drivers/char/watchdog/w83877f_wdt.c | 155 +++++++++++++++++++++++++-----------
2 files changed, 113 insertions(+), 46 deletions(-)

through these ChangeSets:

<wim@iguana.be> (03/08/09 1.1142)
[WATCHDOG] w83877f_wdt patch

cleanup comments and trailing spaces
eliminate extra spin_unlock
add KERN_* tags to printks
added extra printk's to report what problem occured

<wim@iguana.be> (03/08/09 1.1143)
[WATCHDOG] w83877f_wdt.c patch2

add CONFIG_WATCHDOG_NOWAYOUT support
changed watchdog_info to correctly reflect what the driver offers
added WDIOC_GETSTATUS, WDIOC_GETBOOTSTATUS and WDIOC_SETOPTIONS ioctls
use module_param

<wim@iguana.be> (03/08/09 1.1144)
[WATCHDOG] w83877f_wdt.c patch3 (add timeout features)

added WDIOC_SETTIMEOUT and WDIOC_GETTIMEOUT ioctls
made timeout (the emulated heartbeat) a module_param
made the keepalive ping an internal subroutine

<wim@iguana.be> (03/08/09 1.1145)
[WATCHDOG] sbc60xxwdt.c patch6

some small clean-ups: do correct errorhandling


The ChangeSets can also be looked at on:
http://linux-watchdog.bkbits.net:8080/linux-2.5-watchdog

For completeness, I added the patches below.

Greetings,
Wim.

================================================================================
diff -Nru a/drivers/char/watchdog/w83877f_wdt.c b/drivers/char/watchdog/w83877f_wdt.c
--- a/drivers/char/watchdog/w83877f_wdt.c Sat Aug 9 18:40:03 2003
+++ b/drivers/char/watchdog/w83877f_wdt.c Sat Aug 9 18:40:03 2003
@@ -1,38 +1,28 @@
/*
- * W83877F Computer Watchdog Timer driver for Linux 2.4.x
+ * W83877F Computer Watchdog Timer driver
*
* Based on acquirewdt.c by Alan Cox,
- * and sbc60xxwdt.c by Jakob Oestergaard <jakob@ostenfeld.dk>
- *
+ * and sbc60xxwdt.c by Jakob Oestergaard <jakob@unthought.net>
+ *
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version
* 2 of the License, or (at your option) any later version.
- *
- * The authors do NOT admit liability nor provide warranty for
- * any of this software. This material is provided "AS-IS" in
+ *
+ * The authors do NOT admit liability nor provide warranty for
+ * any of this software. This material is provided "AS-IS" in
* the hope that it may be useful for others.
*
* (c) Copyright 2001 Scott Jennings <linuxdrivers@oro.net>
*
* 4/19 - 2001 [Initial revision]
* 9/27 - 2001 Added spinlocking
- *
- *
- * Theory of operation:
- * A Watchdog Timer (WDT) is a hardware circuit that can
- * reset the computer system in case of a software fault.
- * You probably knew that already.
- *
- * Usually a userspace daemon will notify the kernel WDT driver
- * via the /proc/watchdog special device file that userspace is
- * still alive, at regular intervals. When such a notification
- * occurs, the driver will usually tell the hardware watchdog
- * that everything is in order, and that the watchdog should wait
- * for yet another little while to reset the system.
- * If userspace fails (RAM error, kernel bug, whatever), the
- * notifications cease to occur, and the hardware watchdog will
- * reset the system (causing a reboot) after the timeout occurs.
+ * 4/12 - 2002 [rob@osinvestor.com] Eliminate extra comments
+ * Eliminate fop_read
+ * Eliminate extra spin_unlock
+ * Added KERN_* tags to printks
+ * 09/8 - 2003 [wim@iguana.be] cleanup of trailing spaces
+ * added extra printk's for startup problems
*
* This WDT driver is different from most other Linux WDT
* drivers in that the driver will ping the watchdog by itself,
@@ -42,6 +32,7 @@
*/

#include <linux/module.h>
+#include <linux/moduleparam.h>
#include <linux/types.h>
#include <linux/timer.h>
#include <linux/jiffies.h>
@@ -57,6 +48,7 @@
#include <asm/system.h>

#define OUR_NAME "w83877f_wdt"
+#define PFX OUR_NAME ": "

#define ENABLE_W83877F_PORT 0x3F0
#define ENABLE_W83877F 0x87
@@ -95,9 +87,9 @@
static void wdt_timer_ping(unsigned long data)
{
/* If we got a heartbeat pulse within the WDT_US_INTERVAL
- * we agree to ping the WDT
+ * we agree to ping the WDT
*/
- if(time_before(jiffies, next_heartbeat))
+ if(time_before(jiffies, next_heartbeat))
{
/* Ping the WDT */
spin_lock(&wdt_spinlock);
@@ -112,11 +104,11 @@
spin_unlock(&wdt_spinlock);

} else {
- printk(OUR_NAME ": Heartbeat lost! Will not ping the watchdog\n");
+ printk(KERN_WARNING PFX "Heartbeat lost! Will not ping the watchdog\n");
}
}

-/*
+/*
* Utility routines
*/

@@ -147,12 +139,12 @@
next_heartbeat = jiffies + WDT_HEARTBEAT;

/* Start the timer */
- timer.expires = jiffies + WDT_INTERVAL;
+ timer.expires = jiffies + WDT_INTERVAL;
add_timer(&timer);

wdt_change(WDT_ENABLE);

- printk(OUR_NAME ": Watchdog timer is now enabled.\n");
+ printk(KERN_INFO PFX "Watchdog timer is now enabled.\n");
}

static void wdt_turnoff(void)
@@ -162,7 +154,7 @@

wdt_change(WDT_DISABLE);

- printk(OUR_NAME ": Watchdog timer is now disabled...\n");
+ printk(KERN_INFO PFX "Watchdog timer is now disabled...\n");
}


@@ -177,7 +169,7 @@
return -ESPIPE;

/* See if we got the magic character */
- if(count)
+ if(count)
{
size_t ofs;

@@ -204,13 +196,11 @@

static int fop_open(struct inode * inode, struct file * file)
{
- switch(minor(inode->i_rdev))
+ switch(minor(inode->i_rdev))
{
case WATCHDOG_MINOR:
/* Just in case we're already talking to someone... */
if(test_and_set_bit(0, &wdt_is_open)) {
- /* Davej: Is this unlock bogus? */
- spin_unlock(&wdt_spinlock);
return -EBUSY;
}
/* Good, fire up the show */
@@ -224,13 +214,13 @@

static int fop_close(struct inode * inode, struct file * file)
{
- if(minor(inode->i_rdev) == WATCHDOG_MINOR)
+ if(minor(inode->i_rdev) == WATCHDOG_MINOR)
{
if(wdt_expect_close)
wdt_turnoff();
else {
del_timer(&timer);
- printk(OUR_NAME ": device file closed unexpectedly. Will not stop the WDT!\n");
+ printk(KERN_CRIT PFX "device file closed unexpectedly. Will not stop the WDT!\n");
}
}
clear_bit(0, &wdt_is_open);
@@ -244,9 +234,9 @@
{
.options = WDIOF_MAGICCLOSE,
.firmware_version = 1,
- .identity = "W83877F"
+ .identity = "W83877F",
};
-
+
switch(cmd)
{
default:
@@ -265,13 +255,13 @@
.write = fop_write,
.open = fop_open,
.release = fop_close,
- .ioctl = fop_ioctl
+ .ioctl = fop_ioctl,
};

static struct miscdevice wdt_miscdev = {
.minor = WATCHDOG_MINOR,
.name = "watchdog",
- .fops = &wdt_fops
+ .fops = &wdt_fops,
};

/*
@@ -281,21 +271,21 @@
static int wdt_notify_sys(struct notifier_block *this, unsigned long code,
void *unused)
{
- if(code==SYS_DOWN || code==SYS_HALT)
+ if(code==SYS_DOWN || code==SYS_HALT)
wdt_turnoff();
return NOTIFY_DONE;
}
-
+
/*
* The WDT needs to learn about soft shutdowns in order to
- * turn the timebomb registers off.
+ * turn the timebomb registers off.
*/
-
+
static struct notifier_block wdt_notifier=
{
.notifier_call = wdt_notify_sys,
.next = NULL,
- .priority = 0
+ .priority = 0,
};

static void __exit w83877f_wdt_unload(void)
@@ -317,9 +307,20 @@
spin_lock_init(&wdt_spinlock);

if (!request_region(ENABLE_W83877F_PORT, 2, "W83877F WDT"))
+ {
+ printk(KERN_ERR PFX "I/O address 0x%04x already in use\n",
+ ENABLE_W83877F_PORT);
+ rc = -EIO;
goto err_out;
+ }
+
if (!request_region(WDT_PING, 1, "W8387FF WDT"))
+ {
+ printk(KERN_ERR PFX "I/O address 0x%04x already in use\n",
+ WDT_PING);
+ rc = -EIO;
goto err_out_region1;
+ }

init_timer(&timer);
timer.function = wdt_timer_ping;
@@ -327,14 +328,22 @@

rc = misc_register(&wdt_miscdev);
if (rc)
+ {
+ printk(KERN_ERR PFX "cannot register miscdev on minor=%d (err=%d)\n",
+ wdt_miscdev.minor, rc);
goto err_out_region2;
+ }

rc = register_reboot_notifier(&wdt_notifier);
if (rc)
+ {
+ printk(KERN_ERR PFX "cannot register reboot notifier (err=%d)\n",
+ rc);
goto err_out_miscdev;
+ }
+
+ printk(KERN_INFO PFX "WDT driver for W83877F initialised.\n");

- printk(KERN_INFO OUR_NAME ": WDT driver for W83877F initialised.\n");
-
return 0;

err_out_miscdev:
diff -Nru a/drivers/char/watchdog/w83877f_wdt.c b/drivers/char/watchdog/w83877f_wdt.c
--- a/drivers/char/watchdog/w83877f_wdt.c Sat Aug 9 18:40:22 2003
+++ b/drivers/char/watchdog/w83877f_wdt.c Sat Aug 9 18:40:22 2003
@@ -21,8 +21,14 @@
* Eliminate fop_read
* Eliminate extra spin_unlock
* Added KERN_* tags to printks
+ * add CONFIG_WATCHDOG_NOWAYOUT support
+ * fix possible wdt_is_open race
+ * changed watchdog_info to correctly reflect what the driver offers
+ * added WDIOC_GETSTATUS, WDIOC_GETBOOTSTATUS
+ * and WDIOC_SETOPTIONS ioctls
* 09/8 - 2003 [wim@iguana.be] cleanup of trailing spaces
* added extra printk's for startup problems
+ * use module_param
*
* This WDT driver is different from most other Linux WDT
* drivers in that the driver will ping the watchdog by itself,
@@ -73,11 +79,20 @@

#define WDT_HEARTBEAT (HZ * 30)

+#ifdef CONFIG_WATCHDOG_NOWAYOUT
+static int nowayout = 1;
+#else
+static int nowayout = 0;
+#endif
+
+module_param(nowayout, int, 0);
+MODULE_PARM_DESC(nowayout, "Watchdog cannot be stopped once started (default=CONFIG_WATCHDOG_NOWAYOUT)");
+
static void wdt_timer_ping(unsigned long);
static struct timer_list timer;
static unsigned long next_heartbeat;
static unsigned long wdt_is_open;
-static int wdt_expect_close;
+static char wdt_expect_close;
static spinlock_t wdt_spinlock;

/*
@@ -168,62 +183,55 @@
if(ppos != &file->f_pos)
return -ESPIPE;

- /* See if we got the magic character */
+ /* See if we got the magic character 'V' and reload the timer */
if(count)
{
- size_t ofs;
-
- /* note: just in case someone wrote the magic character
- * five months ago... */
- wdt_expect_close = 0;
-
- /* now scan */
- for(ofs = 0; ofs != count; ofs++)
+ if (!nowayout)
{
- char c;
- if (get_user(c, buf + ofs))
- return -EFAULT;
- if (c == 'V')
- wdt_expect_close = 1;
+ size_t ofs;
+
+ /* note: just in case someone wrote the magic character
+ * five months ago... */
+ wdt_expect_close = 0;
+
+ /* scan to see wether or not we got the magic character */
+ for(ofs = 0; ofs != count; ofs++)
+ {
+ char c;
+ if (get_user(c, buf + ofs))
+ return -EFAULT;
+ if (c == 'V')
+ wdt_expect_close = 42;
+ }
}

/* someone wrote to us, we should restart timer */
next_heartbeat = jiffies + WDT_HEARTBEAT;
- return 1;
- };
- return 0;
+ }
+ return count;
}

static int fop_open(struct inode * inode, struct file * file)
{
- switch(minor(inode->i_rdev))
- {
- case WATCHDOG_MINOR:
- /* Just in case we're already talking to someone... */
- if(test_and_set_bit(0, &wdt_is_open)) {
- return -EBUSY;
- }
- /* Good, fire up the show */
- wdt_startup();
- return 0;
+ /* Just in case we're already talking to someone... */
+ if(test_and_set_bit(0, &wdt_is_open))
+ return -EBUSY;

- default:
- return -ENODEV;
- }
+ /* Good, fire up the show */
+ wdt_startup();
+ return 0;
}

static int fop_close(struct inode * inode, struct file * file)
{
- if(minor(inode->i_rdev) == WATCHDOG_MINOR)
- {
- if(wdt_expect_close)
- wdt_turnoff();
- else {
- del_timer(&timer);
- printk(KERN_CRIT PFX "device file closed unexpectedly. Will not stop the WDT!\n");
- }
+ if(wdt_expect_close == 42)
+ wdt_turnoff();
+ else {
+ del_timer(&timer);
+ printk(KERN_CRIT PFX "device file closed unexpectedly. Will not stop the WDT!\n");
}
clear_bit(0, &wdt_is_open);
+ wdt_expect_close = 0;
return 0;
}

@@ -232,7 +240,7 @@
{
static struct watchdog_info ident=
{
- .options = WDIOF_MAGICCLOSE,
+ .options = WDIOF_KEEPALIVEPING | WDIOF_MAGICCLOSE,
.firmware_version = 1,
.identity = "W83877F",
};
@@ -243,9 +251,31 @@
return -ENOIOCTLCMD;
case WDIOC_GETSUPPORT:
return copy_to_user((struct watchdog_info *)arg, &ident, sizeof(ident))?-EFAULT:0;
+ case WDIOC_GETSTATUS:
+ case WDIOC_GETBOOTSTATUS:
+ return put_user(0, (int *)arg);
case WDIOC_KEEPALIVE:
next_heartbeat = jiffies + WDT_HEARTBEAT;
return 0;
+ case WDIOC_SETOPTIONS:
+ {
+ int new_options, retval = -EINVAL;
+
+ if(get_user(new_options, (int *)arg))
+ return -EFAULT;
+
+ if(new_options & WDIOS_DISABLECARD) {
+ wdt_turnoff();
+ retval = 0;
+ }
+
+ if(new_options & WDIOS_ENABLECARD) {
+ wdt_startup();
+ retval = 0;
+ }
+
+ return retval;
+ }
}
}

@@ -342,7 +372,8 @@
goto err_out_miscdev;
}

- printk(KERN_INFO PFX "WDT driver for W83877F initialised.\n");
+ printk(KERN_INFO PFX "WDT driver for W83877F initialised. (nowayout=%d)\n",
+ nowayout);

return 0;

diff -Nru a/drivers/char/watchdog/sbc60xxwdt.c b/drivers/char/watchdog/sbc60xxwdt.c
--- a/drivers/char/watchdog/sbc60xxwdt.c Sat Aug 9 18:41:00 2003
+++ b/drivers/char/watchdog/sbc60xxwdt.c Sat Aug 9 18:41:00 2003
@@ -31,6 +31,7 @@
* made timeout (the emulated heartbeat) a module_param
* made the keepalive ping an internal subroutine
* made wdt_stop and wdt_start module params
+ * added extra printk's for startup problems
* added MODULE_AUTHOR and MODULE_DESCRIPTION info
*
*
@@ -239,7 +240,7 @@
switch(cmd)
{
default:
- return -ENOTTY;
+ return -ENOIOCTLCMD;
case WDIOC_GETSUPPORT:
return copy_to_user((struct watchdog_info *)arg, &ident, sizeof(ident))?-EFAULT:0;
case WDIOC_GETSTATUS:
@@ -364,6 +365,7 @@
{
printk(KERN_ERR PFX "I/O address 0x%04x already in use\n",
wdt_stop);
+ rc = -EIO;
goto err_out_region1;
}
}
-
To unsubscribe from this list: send the line "unsubscribe linux-kernel" in
the body of a message to majordomo@vger.kernel.org
More majordomo info at http://vger.kernel.org/majordomo-info.html
Please read the FAQ at http://www.tux.org/lkml/
\
 
 \ /
  Last update: 2005-03-22 13:47    [W:0.161 / U:0.056 seconds]
©2003-2020 Jasper Spaans|hosted at Digital Ocean and TransIP|Read the blog|Advertise on this site
diff -Nru a/drivers/char/watchdog/w83877f_wdt.c b/drivers/char/watchdog/w83877f_wdt.c
--- a/drivers/char/watchdog/w83877f_wdt.c Sat Aug 9 18:40:41 2003
+++ b/drivers/char/watchdog/w83877f_wdt.c Sat Aug 9 18:40:41 2003
@@ -24,11 +24,13 @@
* add CONFIG_WATCHDOG_NOWAYOUT support
* fix possible wdt_is_open race
* changed watchdog_info to correctly reflect what the driver offers
* added WDIOC_GETSTATUS, WDIOC_GETBOOTSTATUS
* and WDIOC_SETOPTIONS ioctls
* 09/8 - 2003 [wim@iguana.be] cleanup of trailing spaces
* added extra printk's for startup problems
* use module_param
+ * made timeout (the emulated heartbeat) a module_param
+ * made the keepalive ping an internal subroutine
*
* This WDT driver is different from most other Linux WDT
* drivers in that the driver will ping the watchdog by itself,
@@ -77,7 +79,11 @@
* char to /dev/watchdog every 30 seconds.
*/

-#define WDT_HEARTBEAT (HZ * 30)
+#define WATCHDOG_TIMEOUT 30 /* 30 sec default timeout */
+static int timeout = WATCHDOG_TIMEOUT; /* in seconds, will be multiplied by HZ to get seconds to wait for a ping */
+module_param(timeout, int, 0);
+MODULE_PARM_DESC(timeout, "Watchdog timeout in seconds. (1<=timeout<=3600, default=" __MODULE_STRING(WATCHDOG_TIMEOUT) ")");
+

#ifdef CONFIG_WATCHDOG_NOWAYOUT
static int nowayout = 1;
@@ -151,7 +157,7 @@

static void wdt_startup(void)
{
- next_heartbeat = jiffies + WDT_HEARTBEAT;
+ next_heartbeat = jiffies + (timeout * HZ);

/* Start the timer */
timer.expires = jiffies + WDT_INTERVAL;
@@ -172,6 +178,11 @@
printk(KERN_INFO PFX "Watchdog timer is now disabled...\n");
}

+static void wdt_keepalive(void)
+{
+ /* user land ping */
+ next_heartbeat = jiffies + (timeout * HZ);
+}

/*
* /dev/watchdog handling
@@ -206,7 +217,7 @@
}

/* someone wrote to us, we should restart timer */
- next_heartbeat = jiffies + WDT_HEARTBEAT;
+ wdt_keepalive();
}
return count;
}
@@ -240,7 +251,7 @@
{
static struct watchdog_info ident=
{
- .options = WDIOF_KEEPALIVEPING | WDIOF_MAGICCLOSE,
+ .options = WDIOF_KEEPALIVEPING | WDIOF_SETTIMEOUT | WDIOF_MAGICCLOSE,
.firmware_version = 1,
.identity = "W83877F",
};
@@ -255,7 +266,7 @@
case WDIOC_GETBOOTSTATUS:
return put_user(0, (int *)arg);
case WDIOC_KEEPALIVE:
- next_heartbeat = jiffies + WDT_HEARTBEAT;
+ wdt_keepalive();
return 0;
case WDIOC_SETOPTIONS:
{
@@ -276,6 +287,22 @@

return retval;
}
+ case WDIOC_SETTIMEOUT:
+ {
+ int new_timeout;
+
+ if(get_user(new_timeout, (int *)arg))
+ return -EFAULT;
+
+ if(new_timeout < 1 || new_timeout > 3600) /* arbitrary upper limit */
+ return -EINVAL;
+
+ timeout = new_timeout;
+ wdt_keepalive();
+ /* Fall through */
+ }
+ case WDIOC_GETTIMEOUT:
+ return put_user(timeout, (int *)arg);
}
}

@@ -336,6 +363,13 @@

spin_lock_init(&wdt_spinlock);

+ if(timeout < 1 || timeout > 3600) /* arbitrary upper limit */
+ {
+ timeout = WATCHDOG_TIMEOUT;
+ printk(KERN_INFO PFX "timeout value must be 1<=x<=3600, using %d\n",
+ timeout);
+ }
+
if (!request_region(ENABLE_W83877F_PORT, 2, "W83877F WDT"))
{
printk(KERN_ERR PFX "I/O address 0x%04x already in use\n",
@@ -372,8 +406,8 @@
goto err_out_miscdev;
}

- printk(KERN_INFO PFX "WDT driver for W83877F initialised. (nowayout=%d)\n",
- nowayout);
+ printk(KERN_INFO PFX "WDT driver for W83877F initialised. timeout=%d sec (nowayout=%d)\n",
+ timeout, nowayout);

return 0;
@@ -24,11 +24,13 @@
* fix possible wdt_is_open race
* changed watchdog_info to correctly reflect what the driver offers
- * added WDIOC_GETSTATUS, WDIOC_GETBOOTSTATUS
- * and WDIOC_SETOPTIONS ioctls
+ * added WDIOC_GETSTATUS, WDIOC_GETBOOTSTATUS, WDIOC_SETTIMEOUT,
+ * WDIOC_GETTIMEOUT, and WDIOC_SETOPTIONS ioctls
* 09/8 - 2003 [wim@iguana.be] cleanup of trailing spaces
* added extra printk's for startup problems
